import ast
import traceback
import warnings

from miniutils import magic_contract

from .stack import DictStack

try:
    import numpy

    num_types = (int, float, numpy.number)
    float_types = (float, numpy.floating)
except ImportError:  # pragma: nocover
    numpy = None
    num_types = (int, float)
    float_types = (float,)

_collapse_map = {
    ast.Add: lambda a, b: a + b,
    ast.Sub: lambda a, b: a - b,
    ast.Mult: lambda a, b: a * b,
    ast.Div: lambda a, b: a / b,
    ast.FloorDiv: lambda a, b: a // b,

    ast.Mod: lambda a, b: a % b,
    ast.Pow: lambda a, b: a ** b,
    ast.LShift: lambda a, b: a << b,
    ast.RShift: lambda a, b: a >> b,
    ast.MatMult: lambda a, b: a @ b,

    ast.BitAnd: lambda a, b: a & b,
    ast.BitOr: lambda a, b: a | b,
    ast.BitXor: lambda a, b: a ^ b,
    ast.And: lambda a, b: a and b,
    ast.Or: lambda a, b: a or b,
    ast.Invert: lambda a: ~a,
    ast.Not: lambda a: not a,

    ast.UAdd: lambda a: a,
    ast.USub: lambda a: -a,

    ast.Eq: lambda a, b: a == b,
    ast.NotEq: lambda a, b: a != b,
    ast.Lt: lambda a, b: a < b,
    ast.LtE: lambda a, b: a <= b,
    ast.Gt: lambda a, b: a > b,
    ast.GtE: lambda a, b: a >= b,
}


@magic_contract
def can_have_side_effect(node, ctxt):
    """
    Checks whether or not copying the given AST node could cause side effects in the resulting function
    :param node: The AST node to be checked
    :type node: AST
    :param ctxt: The environment stack to use when running the check
    :type ctxt: DictStack
    :return: Whether or not duplicating this node could cause side effects
    :rtype: bool
    """
    if isinstance(node, ast.AST):
        # print("Can {} have side effects?".format(node))
        if isinstance(node, ast.Call):
            # print("  Yes!")
            return True
        else:
            for field, old_value in ast.iter_fields(node):
                if isinstance(old_value, list):
                    return any(can_have_side_effect(n, ctxt) for n in old_value if isinstance(n, ast.AST))
                elif isinstance(old_value, ast.AST):
                    return can_have_side_effect(old_value, ctxt)
                else:
                    # print("  No!")
                    return False
    else:
        return False


@magic_contract
def constant_iterable(node, ctxt, avoid_side_effects=True):
    """
    If the given node is a known iterable of some sort, return the list of its elements.
    :param node: The AST node to be checked
    :type node: AST
    :param ctxt: The environment stack to use when running the check
    :type ctxt: DictStack
    :param avoid_side_effects: Whether or not to avoid unwrapping side effect-causing AST nodes
    :type avoid_side_effects: bool
    :return: The iterable if possible, else None
    :rtype: iterable|None
    """

    # TODO: Support zipping
    # TODO: Support sets/dicts?
    # TODO: Support for reversed, enumerate, etc.
    # TODO: Support len, in, etc.
    # Check for range(*constants)
    def wrap(return_node, name, idx):
        if not avoid_side_effects:
            return return_node
        if can_have_side_effect(return_node, ctxt):
            return ast.Subscript(name, ast.Index(idx))
        return make_ast_from_literal(return_node)

    if isinstance(node, ast.Call):
        if resolve_name_or_attribute(node.func, ctxt) == range:
            args = [resolve_literal(arg, ctxt) for arg in node.args]
            if all(isinstance(arg, ast.Num) for arg in args):
                return [ast.Num(n) for n in range(*[arg.n for arg in args])]

        return None
    elif isinstance(node, (ast.List, ast.Tuple)):
        return [resolve_literal(e, ctxt) for e in node.elts]
        # return [_resolve_name_or_attribute(e, ctxt) for e in node.elts]
    # Can't yet support sets and lists, since you need to compute what the unique values would be
    # elif isinstance(node, ast.Dict):
    #     return node.keys
    elif isinstance(node, (ast.Name, ast.Attribute, ast.NameConstant)):
        res = resolve_name_or_attribute(node, ctxt)
        # print("Trying to resolve '{}' as list, got {}".format(astor.to_source(node), res))
        if isinstance(res, ast.AST) and not isinstance(res, (ast.Name, ast.Attribute, ast.NameConstant)):
            res = constant_iterable(res, ctxt)
        if not isinstance(res, ast.AST):
            try:
                if hasattr(res, 'items'):
                    return dict([(k, wrap(make_ast_from_literal(v), node, k)) for k, v in res.items()])
                else:
                    return [wrap(make_ast_from_literal(res_node), node, i) for i, res_node in enumerate(res)]
            except TypeError:
                pass
    return None


# @magic_contract
def constant_dict(node, ctxt):
    if isinstance(node, (ast.Name, ast.NameConstant, ast.Attribute)):
        res = resolve_name_or_attribute(node, ctxt)
        if hasattr(res, 'items'):
            return dict(res.items())
    return None


@magic_contract
def resolve_name_or_attribute(node, ctxt):
    """
    If the given name of attribute is defined in the current context, return its value. Else, returns the node
    :param node: The node to try to resolve
    :type node: AST
    :param ctxt: The environment stack to use when running the check
    :type ctxt: DictStack
    :return: The object if the name was found, else the original node
    :rtype: *
    """
    if isinstance(node, ast.Name):
        if node.id in ctxt:
            try:
                return ctxt[node.id]
            except KeyError:
                # This occurs if we know that the name was assigned, but we don't know what to... just return the node
                return node
        else:
            return node
    elif isinstance(node, ast.NameConstant):
        return node.value
    elif isinstance(node, ast.Attribute):
        base_obj = resolve_name_or_attribute(node.value, ctxt)
        if not isinstance(base_obj, ast.AST):
            return getattr(base_obj, node.attr, node)
        else:
            return node
    else:
        return node


@magic_contract
def make_ast_from_literal(lit):
    """
    Converts literals into their AST equivalent
    :param lit: The literal to attempt to turn into an AST
    :type lit: *
    :return: The AST version of the literal, or the original AST node if one was given
    :rtype: *
    """
    if isinstance(lit, ast.AST):
        return lit
    elif isinstance(lit, (list, tuple)):
        res = [make_ast_from_literal(e) for e in lit]
        tp = ast.List if isinstance(lit, list) else ast.Tuple
        return tp(elts=res)
    elif isinstance(lit, num_types):
        if isinstance(lit, float_types):
            lit2 = float(lit)
        else:
            lit2 = int(lit)
        if lit2 != lit:
            raise AssertionError("({}){} != ({}){}".format(type(lit), lit, type(lit2), lit2))
        return ast.Num(lit2)
    elif isinstance(lit, str):
        return ast.Str(lit)
    elif isinstance(lit, bool):
        return ast.NameConstant(lit)
    else:
        # warnings.warn("'{}' of type {} is not able to be made into an AST node".format(lit, type(lit)))
        return lit


@magic_contract
def is_wrappable(lit):
    """
    Checks if the given object either is or can be made into a known AST node
    :param lit: The object to try to wrap
    :type lit: *
    :return: Whether or not this object can be wrapped as an AST node
    :rtype: bool
    """
    return isinstance(make_ast_from_literal(lit), ast.AST)


@magic_contract
def _resolve_literal(node, ctxt):
    """
    Collapses literal expressions. Returns literals if they're available, AST nodes otherwise
    :param node: The AST node to be checked
    :type node: AST
    :param ctxt: The environment stack to use when running the check
    :type ctxt: DictStack
    :return: The given AST node with literal operations collapsed as much as possible
    :rtype: *
    """
    # try:
    #     print("Trying to collapse {}".format(astor.to_source(node)))
    # except:
    #     print("Trying to collapse (source not possible) {}".format(astor.dump_tree(node)))

    if isinstance(node, (ast.Name, ast.Attribute, ast.NameConstant)):
        res = resolve_name_or_attribute(node, ctxt)
        if isinstance(res, ast.AST) and not isinstance(res, (ast.Name, ast.Attribute, ast.NameConstant)):
            new_res = _resolve_literal(res, ctxt)
            if is_wrappable(new_res):
                # print("{} can be replaced by more specific literal {}".format(res, new_res))
                res = new_res
            # else:
            #     print("{} is an AST node, but can't safely be made more specific".format(res))
        return res
    elif isinstance(node, ast.Num):
        return node.n
    elif isinstance(node, ast.Str):
        return node.s
    elif isinstance(node, ast.Index):
        return _resolve_literal(node.value, ctxt)
    elif isinstance(node, (ast.Slice, ast.ExtSlice)):
        raise NotImplementedError()
    elif isinstance(node, ast.Subscript):
        # print("Attempting to subscript {}".format(astor.to_source(node)))
        lst = constant_iterable(node.value, ctxt)
        # print("Can I subscript {}?".format(lst))
        if lst is None:
            return node
        slc = _resolve_literal(node.slice, ctxt)
        # print("Getting subscript at {}".format(slc))
        if isinstance(slc, ast.AST):
            return node
        # print("Value at {}[{}] = {}".format(lst, slc, lst[slc]))
        val = lst[slc]
        if isinstance(val, ast.AST):
            new_val = _resolve_literal(val, ctxt)
            if is_wrappable(new_val):
                # print("{} can be replaced by more specific literal {}".format(val, new_val))
                val = new_val
        #     else:
        #         print("{} is an AST node, but can't safely be made more specific".format(val))
        # print("Final value at {}[{}] = {}".format(lst, slc, val))
        return val
    elif isinstance(node, ast.UnaryOp):
        operand = _resolve_literal(node.operand, ctxt)
        if isinstance(operand, ast.AST):
            return node
        else:
            try:
                return _collapse_map[type(node.op)](operand)
            except:
                warnings.warn(
                    "Unary op collapse failed. Collapsing skipped, but executing this function will likely fail."
                    " Error was:\n{}".format(traceback.format_exc()))
    elif isinstance(node, ast.BinOp):
        left = _resolve_literal(node.left, ctxt)
        right = _resolve_literal(node.right, ctxt)
        # print("({} {})".format(repr(node.op), ", ".join(repr(o) for o in operands)))
        lliteral = not isinstance(left, ast.AST)
        rliteral = not isinstance(right, ast.AST)
        if lliteral and rliteral:
            # print("Both operands {} and {} are literals, attempting to collapse".format(left, right))
            try:
                return _collapse_map[type(node.op)](left, right)
            except:
                warnings.warn(
                    "Binary op collapse failed. Collapsing skipped, but executing this function will likely fail."
                    " Error was:\n{}".format(traceback.format_exc()))
                return node
        else:
            left = make_ast_from_literal(left)
            left = left if isinstance(left, ast.AST) else node.left

            right = make_ast_from_literal(right)
            right = right if isinstance(right, ast.AST) else node.right
            # print("Attempting to combine {} and {} ({} op)".format(left, right, node.op))
            return ast.BinOp(left=left, right=right, op=node.op)
    elif isinstance(node, ast.Compare):
        operands = [_resolve_literal(o, ctxt) for o in [node.left] + node.comparators]
        if all(not isinstance(opr, ast.AST) for opr in operands):
            return all(_collapse_map[type(cmp_func)](operands[i - 1], operands[i])
                       for i, cmp_func in zip(range(1, len(operands)), node.ops))
        else:
            return node
    else:
        return node


@magic_contract
def resolve_literal(node, ctxt, give_raw_result=False):
    """
    Collapse literal expressions in the given node. Returns the node with the collapsed literals
    :param node: The AST node to be checked
    :type node: AST
    :param ctxt: The environment stack to use when running the check
    :type ctxt: DictStack
    :return: The given AST node with literal operations collapsed as much as possible
    :rtype: *
    """
    result = _resolve_literal(node, ctxt)
    if give_raw_result:
        return result
    result = make_ast_from_literal(result)
    if not isinstance(result, ast.AST):
        return node
    return result



import ast
import inspect

import astor
from miniutils.magic_contract import safe_new_contract


def _is_iterable(x):
    try:
        iter(x)
        return True
    except Exception:
        return False


safe_new_contract('iterable', _is_iterable)
safe_new_contract('literal', 'int|float|str|bool|tuple|list|None')
for name, tp in inspect.getmembers(ast, inspect.isclass):
    safe_new_contract(name, tp)

# Astor tries to get fancy by failing nicely, but in doing so they fail when traversing non-AST type node properties.
#  By deleting this custom handler, it'll fall back to the default ast visit pattern, which skips these missing
# properties. Everything else seems to be implemented, so this will fail silently if it hits an AST node that isn't
# supported but should be.
try:
    del astor.node_util.ExplicitNodeVisitor.visit
except AttributeError:
    # visit isn't defined in this version of astor
    pass

from .stack import DictStack
from .resolve import *
from .transformer import *



# import ast
import copy
import inspect
import sys
import tempfile
import textwrap

import astor
from miniutils.opt_decorator import optional_argument_decorator

from .resolve import *
from .stack import DictStack


@magic_contract
def function_ast(f):
    """
    Returns ast for the given function. Gives a tuple of (ast_module, function_body, function_file
    :param f: The function to parse
    :type f: Callable
    :return: The relevant AST code: A module including only the function definition; the func body; the func file
    :rtype: tuple(Module, list(AST), str)
    """
    try:
        f_file = sys.modules[f.__module__].__file__
    except (KeyError, AttributeError):
        f_file = ''

    root = ast.parse(textwrap.dedent(inspect.getsource(f)), f_file)
    return root, root.body[0].body, f_file


@magic_contract
def _assign_names(node):
    """
    Gets names from a assign-to tuple in flat form, just to know what's affected
    "x=3" -> "x"
    "a,b=4,5" -> ["a", "b"]
    "(x,(y,z)),(a,) = something" -> ["x", "y", "z", "a"]

    :param node: The AST node to resolve to a list of names
    :type node: AST
    :return: The flattened list of names referenced in this node
    :rtype: iterable
    """
    if isinstance(node, ast.Name):
        yield node.id
    elif isinstance(node, ast.Tuple):
        for e in node.elts:
            yield from _assign_names(e)
    elif isinstance(node, ast.Subscript):
        yield from _assign_names(node.value)


class DebugTransformerMixin:  # pragma: nocover
    def visit(self, node):
        orig_node_code = astor.to_source(node).strip()
        print("Starting to visit >> {} <<".format(orig_node_code))

        new_node = super().visit(node)

        try:
            if new_node is None:
                print("Deleted >>> {} <<<".format(orig_node_code))
            elif isinstance(new_node, ast.AST):
                print("Converted >>> {} <<< to >>> {} <<<".format(orig_node_code, astor.to_source(new_node).strip()))
            elif isinstance(new_node, list):
                print("Converted >>> {} <<< to [[[ {} ]]]".format(orig_node_code, ", ".join(
                    astor.to_source(n).strip() for n in new_node)))
        except Exception as ex:
            raise AssertionError("Failed on {} >>> {}".format(orig_node_code, astor.dump_tree(new_node))) from ex
            # print("Failed on {} >>> {}".format(astor.dump_tree(orig_node), astor.dump_tree(new_node)))
            # return orig_node

        return new_node


class TrackedContextTransformer(ast.NodeTransformer):
    def __init__(self, ctxt=None):
        self.ctxt = ctxt or DictStack()
        super().__init__()

    def visit_many(self, nodes):
        for n in nodes:
            n = self.visit(n)
            if n is not None:
                if isinstance(n, ast.AST):
                    yield n
                else:
                    yield from n

    def visit_Assign(self, node):
        node.value = self.visit(node.value)
        erase_targets = True
        # print(node.value)
        # TODO: Support tuple assignments
        if len(node.targets) == 1:
            if isinstance(node.targets[0], ast.Name):
                nvalue = copy.deepcopy(node.value)
                var = node.targets[0].id
                val = constant_iterable(nvalue, self.ctxt)
                if val is not None:
                    # print("Setting {} = {}".format(var, val))
                    self.ctxt[var] = val
                else:
                    val = resolve_literal(nvalue, self.ctxt)
                    # print("Setting {} = {}".format(var, val))
                    self.ctxt[var] = val
                erase_targets = False
            # elif isinstance(node.targets[0], ast.Subscript):
            #     targ = node.targets[0]
            #     iterable = constant_iterable(targ.value, self.ctxt, False)
            #     if iterable is None:
            #         iterable = constant_dict(targ.value, self.ctxt)
            #     if iterable is None:
            #         return node
            #     key = resolve_literal(targ.slice, self.ctxt)
            #     if isinstance(key, ast.AST):
            #         return node
            #
            #     nvalue = copy.deepcopy(node.value)
            #     val = constant_iterable(nvalue, self.ctxt)
            #     warnings.warn("Iterable assignment not fully implemented yet...")
            #     if val is not None:
            #         # print("Setting {} = {}".format(var, val))
            #         iterable[key] = val
            #     else:
            #         val = resolve_literal(nvalue, self.ctxt)
            #         # print("Setting {} = {}".format(var, val))
            #         iterable[key] = val
            #     erase_targets = False

        if erase_targets:
            for targ in node.targets:
                for assgn in _assign_names(targ):
                    self.ctxt[assgn] = None
        return node

    def visit_AugAssign(self, node):
        for assgn in _assign_names(node.target):
            self.ctxt[assgn] = None
        return super().generic_visit(node)

    def visit_Delete(self, node):
        for targ in node.targets:
            for assgn in _assign_names(targ):
                del self.ctxt[assgn]
        return super().generic_visit(node)

    def visit_FunctionDef(self, node):
        self.ctxt.push({}, False)
        node.body = list(self.visit_many(node.body))
        self.ctxt.pop()
        return self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node):
        self.ctxt.push({}, False)
        node.body = list(self.visit_many(node.body))
        self.ctxt.pop()
        return self.generic_visit(node)

    def visit_ClassDef(self, node):
        self.ctxt.push({}, False)
        node.body = list(self.visit_many(node.body))
        self.ctxt.pop()
        return self.generic_visit(node)


def make_function_transformer(transformer_type, name, description, **transformer_kwargs):
    @optional_argument_decorator
    @magic_contract
    def transform(return_source=False, save_source=True, function_globals=None, **kwargs):
        """
        :param return_source: Returns the transformed function's source code instead of compiling it
        :type return_source: bool
        :param save_source: Saves the function source code to a tempfile to make it inspectable
        :type save_source: bool
        :param function_globals: Overridden global name assignments to use when processing the function
        :type function_globals: dict|None
        :param kwargs: Any other environmental variables to provide during unrolling
        :type kwargs: dict
        :return: The transformed function, or its source code if requested
        :rtype: Callable
        """

        @magic_contract(f='Callable', returns='Callable|str')
        def inner(f):
            f_mod, f_body, f_file = function_ast(f)
            # Grab function globals
            glbls = f.__globals__
            # Grab function closure variables
            if isinstance(f.__closure__, tuple):
                glbls.update({k: v.cell_contents for k, v in zip(f.__code__.co_freevars, f.__closure__)})
            # Apply manual globals override
            if function_globals is not None:
                glbls.update(function_globals)
            # print({k: v for k, v in glbls.items() if k not in globals()})
            trans = transformer_type(DictStack(glbls, kwargs), **transformer_kwargs)
            f_mod.body[0].decorator_list = []
            f_mod = trans.visit(f_mod)
            # print(astor.dump_tree(f_mod))
            if return_source or save_source:
                try:
                    source = astor.to_source(f_mod)
                except ImportError:  # pragma: nocover
                    raise ImportError("miniutils.pragma.{name} requires 'astor' to be installed to obtain source code"
                                      .format(name=name))
                except Exception as ex:  # pragma: nocover
                    raise RuntimeError(astor.dump_tree(f_mod)) from ex
            else:
                source = None

            if return_source:
                return source
            else:
                f_mod = ast.fix_missing_locations(f_mod)
                if save_source:
                    temp = tempfile.NamedTemporaryFile('w', delete=True)
                    f_file = temp.name
                exec(compile(f_mod, f_file, 'exec'), glbls)
                func = glbls[f_mod.body[0].name]
                if save_source:
                    func.__tempfile__ = temp
                    temp.write(source)
                    temp.flush()
                return func

        return inner

    transform.__name__ = name
    transform.__doc__ = '\n'.join([description, transform.__doc__])
    return transform
